########################################################################
# Klasse fr die Scans #################################################
########################################################################
package loadscan;

use strict;
use warnings;
use Gtk2;
use Data::Dumper;

# gibt bei den rohdaten an welches Bild genommen werden soll
# gltige Werte sind "forward" und "backward"
# ist ein ungltiger wert gesetzt wird 'forward' genommen!
our $scan_direction='forward';
# setzt man ber $loadscan::scan_direction=''
# das ist eine "Klassenvariable"!


# erzeuge Scanobjekt
# bergeben werden muss:
#  - die datei (ohen bergeben Datei funtionieren viele Methoden nicht!)
#  - der contrast (0 bedeutet autokontrast) default ist 0
#  - Nachkommastellen fr die Verteilungskurve default ist 3
sub new
{
  my $class=shift;
  my $file=shift;
  my $contr=shift || 0;
  my $stellen=shift || 3;
  my $self={};
  $self->{matrix}=undef;
  $self->{niveaulist}=undef;
  $self->{width}=1;
  $self->{height}=1;
  $self->{deep}=1;
  $self->{real_height}=0;
  $self->{real_width}=0;
  $self->{real_width_offset}=0;
  $self->{real_heigth_offset}=0;

  $self->{stepping_width}=0.3;
  $self->{stepping_height}=0.3;
  $self->{stepping_deep}=0.3;

  $self->{real_max_deep}=0;
  $self->{real_min_deep}=0;
  $self->{unit_height}='nm?';
  $self->{unit_width}='nm?';
  $self->{unit_deep}='nm?';
  $self->{middle_deep}=-1;
  $self->{contrast}=$contr;
  $self->{file}='';
  $self->{file_b}='';
  $self->{file_f}='';
  $self->{stellen}=$stellen;
  $self->{pixbuf}=undef;
  bless($self,$class);
  if($file)
  { return undef unless($self->read($file)); }
  return $self;
}

#=======================================================================
# GETTER

#-----------------------------------------------------------------------
#aktualler Dateiname
sub get_file
{ return shift()->{file}; }

#-----------------------------------------------------------------------
# Kontrast neu setzen
sub set_contrast
{ $_[0]->{contrast}=$_[1]; }

#-----------------------------------------------------------------------
# Nachkommastellen fr die Verteilungskurve setzen
# dazu wird das Bild noch einmal geladen!
sub set_stellen
{
  my $self=shift;
  my $val=shift;

  if($self->{stellen} != $val)
  {
    $self->{stellen} = $val;
    # damit das Bild wirklich neu geladen wird
    my $file=$self->{file};
    $self->{file}='';
    $self->read($file) if($file);
  }
}

#-----------------------------------------------------------------------
# Nachkommastellen fr die Verteilungskurve lesen
sub get_stellen
{ return shift()->{stellen}; }

#-----------------------------------------------------------------------
# Hhe des Gescanten bildes in pixel
sub get_height
{ return shift()->{height}+1; }

#-----------------------------------------------------------------------
# Breite des gescanten Bildes in Pixel
sub get_width
{ return shift()->{width}+1; }

#-----------------------------------------------------------------------
# Hhe des Bildes in nm
sub get_real_height
{ return shift()->{real_height}; }

#-----------------------------------------------------------------------
#Breite des Bildes in nm
sub get_real_width
{ return shift()->{real_width}; }

#-----------------------------------------------------------------------
# grte Hhe des bildes in nm (z-Achse)
sub get_real_max_deep
{ return shift()->{real_max_deep}; }

#-----------------------------------------------------------------------
# Geringste Hhe des Bildes in nm
sub get_real_min_deep
{ return shift()->{real_min_deep}; }

#-----------------------------------------------------------------------
# Mittelwert an Hhe in nm
sub get_real_middle_deep
{ return shift()->{middle_deep}; }

#-----------------------------------------------------------------------
# geringster xyz Wert des Bildes
sub get_all_real_min
{ return (0,0,shift()->{real_min_deep}); }

#-----------------------------------------------------------------------
# grter xyz wert des Bildes
sub get_all_real_max
{
  my $self=shift;
  return ($self->{real_width},$self->{real_height},$self->{real_max_deep});
}

#-----------------------------------------------------------------------
# Schrittweite der Breite des scans in nm
sub get_stepping_width
{ return shift->{stepping_width}; }

#-----------------------------------------------------------------------
# Schrittweite der Lnge des scans in nm
sub get_stepping_height
{ return shift->{stepping_height}; }

#-----------------------------------------------------------------------
# Schrittweite der Hhe des scans in nm
sub get_stepping_deep
{ return shift->{stepping_deep}; }

#-----------------------------------------------------------------------
# Nreile Lnge Hhe Schrittweiten in nm
sub get_stepping
{
  my $self=shift;
  return ($self->get_stepping_width(),
          $self->get_stepping_height(),
          $self->get_stepping_deep());
}

#-----------------------------------------------------------------------
# Grenangaben lesen
sub get_dimensions
{
  my $self=shift;
  return ($self->{unit_width},$self->{unit_height},$self->{unit_deep});
}

#-----------------------------------------------------------------------
# Datenmatrix auslesen
# ein zweidimensionales Array das alle gelesen Werte enthlt
sub get_matrix
{ return shift()->{matrix}; }

#-----------------------------------------------------------------------
# Stck aus dem Scan herausschneiden
# Wenn der Berich auerhealb des Scans liegt,
# dann soll der Rest gespiegelt werden
sub get_matrix_part
{
  my $self=shift;
  return undef unless(defined($self->{matrix}));

  my $minx=shift;
  my $maxx=shift;
  my $miny=shift;
  my $maxy=shift;

  # Werte tauschen wenn wenn min>max
  ($maxx,$minx)=($minx,$maxx) if($minx>$maxx);
  ($maxy,$miny)=($miny,$maxy) if($miny>$maxy);

  my $w=$maxx-$minx;
  my $h=$maxy-$miny;

  my @ret;
  for my $y (0..$h)
  {
    for my $x (0..$w)
    {
      my $px=$minx+$x;
      my $py=$miny+$y;

      # am Rand spiegeln
      $px=abs($px) if($px < 0);
      $px=$self->{width}-($px-$self->{width}) if($px > $self->{width});
      $py=abs($py) if($py < 0);
      $py=$self->{height}-($py-$self->{height}) if($py > $self->{height});

      $ret[$y][$x]=$self->{matrix}->[$py]->[$px];
    }
  }
  return \@ret;
}

#-----------------------------------------------------------------------
# hhe an einer bestimmten position
sub get_deep_at_pos
{
  my $self=shift;
  my ($x,$y)=@_;
  return undef unless(defined($self->{matrix}) &&
                      defined($self->{matrix}->[$y]) &&
                      defined($self->{matrix}->[$y]->[$x]));
  return $self->{matrix}->[$y]->[$x];
}

#-----------------------------------------------------------------------
# gibt eine Hashref zurck,
# die alle ber die stellen nivelieren Punkte auflistet
sub get_nieaulist
{ return shift()->{niveaulist}; }


#-----------------------------------------------------------------------
# Bitestream des scans als rgb-Werte folge
# Jeder Wert ist 8 Bit 3 Werte ergeben einen Bildpunkt.
# Das macht eine Farbtiefe von 24Bit
#
# Wenn keine Werte bergeben werden wird der gesammte Scan geliefert
sub get_raw_image_data
{
  my $self=shift;

  my ($minx,$maxx,$miny,$maxy)=@_;

  # Wenn noch kein Bild geladen wurde kann auch nichts zurckgeben werden
  return undef if(!defined($self->{matrix}));

  $minx=0 unless(defined($minx));
  $miny=0 unless(defined($miny));
  $maxx=$self->{width} unless(defined($maxx));
  $maxy=$self->{height} unless(defined($maxy));

  ($minx,$maxx)=($maxx,$minx) if($minx>$maxx);
  ($miny,$maxy)=($maxy,$miny) if($miny>$maxy);

  my $multiply=255;
  my $nivelize=0;
  # Wenn der Kontrast =0 ist
  # dann soll der Kontrastausgleich automatisch passieren
  if($self->{contrast}==0)
  {
    $multiply=255/($self->{real_max_deep}-$self->{real_min_deep});
    $nivelize=$self->{real_min_deep};
  }
  # Voreingestllten Kontrast nutzen
  elsif($self->{contrast}>0)
  { $multiply=255/($self->{contrast}*$self->{middle_deep}); }
  # Dynamischer Kontrastausgleich bei negetiven Wert
  else
  {
    my $kontrast=abs($self->{contrast});
    # ist noch nicht implemetiert
    # darum nutzen wir erstmal den klassichen Kontrastausgleich
    # Passen aber das Nullevel an.
    $multiply=255/($kontrast*$self->{middle_deep});
    $nivelize=$self->{real_min_deep};
  }

  # Daten konvertieren
  my $data;
  for my $y ($miny..$maxy)
  {
    for my $x ($minx..$maxx)
    {
      # am Rand spiegeln
      $x=abs($x) if($x < 0);
      $x=$self->{width}-($x-$self->{width}) if($x > $self->{width});
      $y=abs($y) if($y < 0);
      $y=$self->{height}-($y-$self->{height}) if($y > $self->{height});

      my $z=$self->{matrix}->[$y]->[$x];
      $z=$self->{real_min_deep} unless($z);

      my $val=($z-$nivelize)*$multiply;
      $val=0 if($val<0);
      $val=255 if($val>255);
      #                 R    G    B
      $data.=pack('C*',$val,$val,$val);
    }
  }
  return $data;
}

#-----------------------------------------------------------------------
# liefert einen Pixelbuffer eines angebenen Ausschittes zurck
sub get_pixbuf_part
{
  my $self=shift;
  return undef unless(defined($self->{matrix}));

  my $minx=shift;
  my $maxx=shift;
  my $miny=shift;
  my $maxy=shift;
  my $w=abs($maxx-$minx);
  my $h=abs($maxy-$miny);
  return Gtk2::Gdk::Pixbuf->new_from_data($self->get_raw_image_data($minx,$maxx,$miny,$maxy),'GDK_COLORSPACE_RGB',0,8,$w+1,$h+1,($w+1)*3)
}

#-----------------------------------------------------------------------
# liefert den Pixelbuffer des erzeugten Scans zurck
sub get_pixbuf
{
  my $self=shift;
  return undef unless(defined($self->{matrix}));

  if(!defined($self->{pixbuf}))
  {
    $self->{pixbuf} = Gtk2::Gdk::Pixbuf->new_from_data($self->get_raw_image_data(0,$self->{width},0,$self->{height}),'GDK_COLORSPACE_RGB',0,8,$self->{width}+1,$self->{height}+1,($self->{width}+1)*3);
  }
  return $self->{pixbuf};
}

#-----------------------------------------------------------------------
# Matrix it mittelwerten
# Ich betutze hier die Graphikfunktionen von GTK2
# Das ist sehr viel schneller als man es in Perl machen knnte
# Man bekommt eine ArrayMatrix zurck oder undef bei Fehler
# Man ber gibt den faktor der skalierung, oder wenn nicht wird 10 angenommen
sub get_scaled_matrix
{
  my $self=shift;
  my $factor=shift;
  $factor=10 unless(defined($factor));
  $factor=abs($factor);
  my $w=int($self->get_width()/$factor);
  my $h=int($self->get_height()/$factor);
  return undef if($w<=0 || $h<=0 );

  my $pixbuf=$self->get_pixbuf();
  return undef if(!defined($pixbuf));

  my $scaled=$pixbuf->scale_simple ($w, $h, 'bilinear');
  my $pixels=$scaled->get_pixels();
  my $matrix=[];
  for my $y (0..$h)
  {
    for my $x (0..$w)
    {
      my $pos=$pixbuf->get_rowstride()*$y+$pixbuf->get_n_channels()*$x;
      my ($r, $g, $b, $a) = unpack("C*", substr($pixels, $pos, $pixbuf->get_n_channels));
      $matrix->[$x]->[$y]=$r;
    }
  }
  return $matrix;
}

#=======================================================================
# SETTER

#-----------------------------------------------------------------------
# Datenstze (scan) lesen
# liefert 1 bei erfolgreichem Lesen und 0 bei Fehler zurck
sub read
{
  my $self=shift;
  my $file=shift;
  $self->{pixbuf}=undef;
  $self->{nivelize}=undef;
  $self->{copy}=undef;
  # Wenn die selbe datei nochmal ausgelesen werden soll,
  # einfach so tun als ob
  return 1 if($self->{file} eq $file);
  $self->{file}=$file;

  # par-Datei an endung erkennen
  return $self->_load_par() if($file=~/\.par$/);

  # ist txt-datei
  return $self->_load_txt();
}

#=======================================================================
# PROZESSOREN

#-----------------------------------------------------------------------
# Die Einzelnen Lines des Scans auf eine Ebene heben/senken
# Es nimmt die meisten punkte einer Hhe
# und setzt diese auf den bergebenen wert.
# Alle anderen Punkte werden relativ dazu angehoben/abgesenkt
# der Defaultwert ist 2 (nm)
sub nivelize_matrix
{
  my $self=shift;
  my $niveau=shift || 2;

  if($self->{nivelize})
  {
    $self->{matrix}=$self->{copy};
    $self->{pixbuf}=undef;
  }

  if( !defined($self->{nivelize}) || $niveau != $self->{nivelize} )
  {
    $self->{pixbuf}=undef;
    $self->{nivelize}=$niveau;
    if($niveau > 0)
    {
      $self->{copy}=[];
      for my $py (0..$#{$self->{matrix}})
      {
        my %points_deep;
        for my $px (0..$#{$self->{matrix}->[$py]})
        {
          my $deep=$self->{matrix}->[$py]->[$px];
          $self->{copy}->[$py]->[$px]=$deep;
          $points_deep{$deep}++;
        }
        my @deeps=sort{$points_deep{$b} <=> $points_deep{$a}}keys(%points_deep);
        my $most=$deeps[0];
        for my $px (0..$#{$self->{matrix}->[$py]})
        {
          my $deep=$self->{matrix}->[$py]->[$px];
          $deep=$deep-$most+$niveau;
          $self->{matrix}->[$py]->[$px]=$deep;
        }
      }
    }
  }
}

########################################################################
# private Methoden #####################################################
########################################################################

# Datenmatix in ein Pixelbuffer umwandeln
sub _matrix_to_pixbuf
{
  my $self=shift;
  my $matrix=shift;

  my $multiply=255;
  my $nivelize=0;
  # Wenn der Kontrast =0 ist
  # dann soll der Kontrastausgleich automatisch passieren
  if($self->{contrast}==0)
  {
    $multiply=255/($self->{real_max_deep}-$self->{real_min_deep});
    $nivelize=$self->{real_min_deep};
  }
  # Voreingestllten Kontrast nutzen
  elsif($self->{contrast}>0)
  { $multiply=255/($self->{contrast}*$self->{middle_deep}); }
  # Dynamischer Kontrastausgleich bei negetiven Wert
  else
  {
    my $kontrast=abs($self->{contrast});
    # ist noch nicht implemetiert
    # darum nutzen wir erstmal den klassichen Kontrastausgleich
    # Passen aber das Nullevel an.
    $multiply=255/($kontrast*$self->{middle_deep});
    $nivelize=$self->{real_min_deep};
  }

  # Daten konvertieren
  my $data;
  for my $y (0..$#$matrix)
  {
    for my $x (0..$#{$matrix->[$y]})
    {
      my $z=$matrix->[$y]->[$x];

      my $val=($z-$nivelize)*$multiply;
      $val=0 if($val<0);
      $val=255 if($val>255);
      #                 R    G    B
      $data.=pack('C*',$val,$val,$val);
    }
  }
  return Gtk2::Gdk::Pixbuf->new_from_data($data,'GDK_COLORSPACE_RGB',0,8,scalar(@$matrix),scalar(@{$matrix->[0]}),@{$matrix->[0]}*3);
}

# parameter datei laden
# ist eine Datei mit der Endung ".par"
sub _load_par
{
  my $self=shift;
  my $fileb='';
  my $filef='';
  $self->{unit_height}='nm';
  $self->{unit_width}='nm';

  return 0 unless(open(my $fh, '<', $self->{file}));
  my $forward=undef;
  my $backward=undef;

  $scan_direction='forward' if($scan_direction ne 'forward' && $scan_direction ne 'backward');

  while(my $line=<$fh>)
  {
    chomp($line);

    # feldDimension auslesen
    $self->{real_width}=$1 if($line=~/Field X Size in nm\s*:\s*(\d+(?:.\d+)?)/);
    $self->{real_height}=$1 if($line=~/Field Y Size in nm\s*:\s*(\d+(?:.\d+)?)/);

    # gre in Schritten
    $self->{width}=$1 if($line=~/Image Size in X\s*:\s*(\d+)/);
    $self->{height}=$1 if($line=~/Image Size in Y\s*:\s*(\d+)/);

    # Schrittweite
    $self->{stepping_width}=$1 if($line=~/Increment X\s*:\s*(\d+(?:.\d+)?)/);
    $self->{stepping_height}=$1 if($line=~/Increment Y\s*:\s*(\d+(?:.\d+)?)/);

    # offset
    #$self->{real_width_offset}=$1 if($line=~/X Offset\s*:\s*([+-]?\d+(?:.\d+)?)/);
    #$self->{real_heigth_offset}=$1 if($line=~/Y Offset\s*:\s*([+-]?\d+(?:.\d+)?)/);

    # eigenschften der Dateien,
    # welche die Daten enthalten
    if($line=~/^\s*Topographic Channel/i)
    {
      my @list=();
      my $cnt=0;
      while(my $bline=<$fh>)
      {
        $bline=~/^\s*(.+?)\s*(?:;.*)?$/s;
        push(@list,$1);
        last if($cnt==8);
        $cnt++;
      }
      #print "@list\n";
      if(@list>=7)
      {
        my $opts={};
        $opts->{min}=$list[1];
        $opts->{max}=$list[2];
        $opts->{raw_min}=$list[3];
        $opts->{raw_max}=$list[4];
        $opts->{stepping}=$list[5];
        $opts->{unit}=$list[6] || 'nm';
        if($list[0] eq 'Forward')
        {
          $filef=$list[7];
          $forward=$opts;
        }
        elsif($list[0] eq 'Backward')
        {
          $fileb=$list[7];
          $backward=$opts;
        }
      }
    }
  }
  close $fh;

  return 0 unless($self->{width});
  return 0 unless($self->{height});

  # TODO:
  # Verbesserung der Pfadermittelung!
  # aktuellen pfad ermitteln (Nur Unixartige systeme!)
  my $dir=__base_path($self->{file});

  my $loaded=0;
  if($filef && $scan_direction eq 'forward')
  {
    $self->{file_f}="$dir/$filef";
    $loaded=$self->_load_raw_data($self->{file_f},$forward)
  }

  if($fileb && $scan_direction eq 'backward')
  {
    $self->{file_b}="$dir/$fileb";
    $loaded=$self->_load_raw_data($self->{file_b},$backward) unless($loaded)
  }

  return $loaded;
}

# rohdaten datei lesen
# wird nur von _load_par benutzt!
sub _load_raw_data
{
  my $self=shift;
  my $file=shift;
  my $opts=shift;

  return 0 if(!$opts or ref($opts) ne 'HASH');
  # vorlufig, die Werte aus der par Datei nehmen
  #$self->{real_max_deep}=$opts->{raw_max};
  #$self->{real_min_deep}=$opts->{raw_min};
  $self->{real_max_deep}=undef;
  $self->{real_min_deep}=undef;
  $self->{deep}=$opts->{max}-$opts->{min};
  $self->{stepping_deep}=$opts->{stepping};
  # Durch runden und ungenauigkeiten kann es zu Abweichungen kommen
  # deshalb werden die Werte intern neu berechnet und gesetzt

  $self->{unit_deep}=$opts->{unit};

  # perl Magie Trick "slurp" ;)
  my $data='';
  {
    local $/=undef;
    open(my $fh,'<',$file) or return 0;
    binmode($fh);
    $data=<$fh>;
    close($fh);
  }
  #my $data=eval{local($/,@ARGV)=(undef,$file); <>};
  #return 0 unless($data);

  my @matrix;
  my $h=$self->{height};
  my $w=$self->{width};

  my $summ_deep=0;
  $self->{niveaulist}={};
  my %zpoints=();

  for my $py (0..($h-1))
  {
    # eine Datenzeile herausscheiden
    my $line=substr($data,$py*$w*2,$w*2);

    # Force little-endian byte-order on signed Short (16bit).
    my @data=unpack('s>*',$line);
    #print "LINE $px\n";

    for my $px (0..$#data)
    {
      my $z=$data[$px];
      # als realwert;
      $z*=$opts->{stepping};

      # Werte auf >=0 setzen
      # anheben/absenken:
      #$z-=$opts->{raw_min};
      # Grte hhe ermitteln
      $self->{real_max_deep}=$z if(!defined($self->{real_max_deep}) || $self->{real_max_deep}<$z);
      # Kleinste hhe ermitteln
      $self->{real_min_deep}=$z if(!defined($self->{real_min_deep}) ||$self->{real_min_deep}>$z);
      # Sollte ansich auch ohne funktionieren, bin nur vorsichtig ...

      # Eine Liste der Hufigkeit der Hhenwerte ermitteln
      # auf ein bestimmte anzahl von nachkommastellen beschrnken:
      my $key=$z.'';
      $key.='.' if($key=~/^\d+$/);
      $key.=('0'x$self->{stellen});
      $key=~s/^(\d+\.\d{$self->{stellen}}).+$/$1/;
      $self->{niveaulist}->{$key}++;

      # zum ermitteln der Schrittweite in z-Richtung
      $zpoints{$z}++;
      # Summe fr den Durchschittswert
      $summ_deep+=$z;

      $self->{matrix}->[$py]->[$px]=$z;
    }
  }

  # Durchschnittswert ermitteln
  $self->{middle_deep}=$summ_deep/($self->{width}*$self->{height});

  #die(Dumper(\%zpoints));
  # Abzahl der Schitte in z-Richtung
  $self->{deep}=scalar(keys(%zpoints));
  # Schrittweite in z-Richtung
  $self->{stepping_deep}=($self->{real_max_deep}-$self->{real_min_deep})/$self->{deep};

  # ist das nur beim Textformat so?
  #@{$self->{matrix}}=reverse(@{$self->{matrix}});
  return 1;
}

# textformat
sub _load_txt
{
  my $self=shift;
  # Auslesen
  return 0 unless(open(my $fh, '<', $self->{file}));

  $self->{real_max_deep}=undef;
  $self->{real_min_deep}=undef;

  my $px=-1;
  my $py=0;
  my $lastx=0;
  my $lasty=0;
  my $summ_deep=0;
  $self->{niveaulist}={};
  my %zpoints=();
  while(my $line=<$fh>)
  {
    # Maeinheiten auslesen
    # wenn eine Zeile nach
    # X[<wert>] Y[<wert>] Z[<Wert>]
    # ausieht dann ist es die Zeile mit den Dimensionen
    if($line=~/^\s*X\[(\w+)\]\s+Y\[(\w+)\]\s+Z\[(\w+)\]/)
    {
      $self->{unit_height}=$1;
      $self->{unit_width}=$2;
      $self->{unit_deep}=$3;
    }
    # Werte Auslesen
    # Eine Zeile sieht so aus:
    # <Zahl mit Punkt/Komma> <Zahl mit Punkt/Komma> <Zahl mit Punkt/Komma>
    elsif(my ($x,$y,$z)=$line=~/^\s*([.,\d]+)\s+([.,\d]+)\s+([.,\d]+)\s*$/)
    {
      # zum ermitteln der Schrittweite in z-Richtung
      $zpoints{$z}++;

      # Eine Liste der Hufigkeit der Hhenwerte ermitteln
      # auf ein bestimmte anzahl von nachkommastellen beschrnken:
      my $key=$z;
      $key.='.' if($key=~/^\d+$/);
      $key.=('0'x$self->{stellen});
      $key=~s/^(\d+\.\d{$self->{stellen}}).+$/$1/;
      $self->{niveaulist}->{$key}++;

      # Grte hhe ermitteln
      $self->{real_max_deep}=$z if(!defined($self->{real_max_deep}) || $self->{real_max_deep}<$z);
      # Kleinste hhe ermitteln
      $self->{real_min_deep}=$z if(!defined($self->{real_min_deep}) ||$self->{real_min_deep}>$z);
      # Summe fr den Durchschittswert
      $summ_deep+=$z;
      # Wenn ein Zeilensprung erfolgt
      # erkennt man daran,
      # dass der Y Wert grer wird un der X Wert kleiner
      if($lastx>$x && $lasty<$y)
      {
        # Breite des Scans in pixlen anpassen
        # Maximale Breite ermitteln
        if($self->{width}<$px)
        { $self->{width}=$px; }

        # Echte Breite des Scans in nm anpassen
        # Maximale Echte breite ermitteln
        if($self->{real_width}<$lastx)
        {
          $self->{real_width}=$lastx;
          #print "NEW real_width=$lastx\n";
        }

        # Zhler zurcksetzen/erhhen
        $px=-1;
        $py++;
      }
      # Nchste Spalte
      $px++;

      # Wert in Matrix setzen
      $self->{matrix}->[$py]->[$px]=$z;

      # Alte Werte sichern zum Vergleich
      $lastx=$x;
      $lasty=$y;
    }
  }
  close($fh);

  $self->{height}=$py;
  $self->{real_height}=$lasty;
  # Durchschnittswert ermitteln
  $self->{middle_deep}=$summ_deep/($self->{width}*$self->{height});
  # Abzahl der Schitte in z-Richtung
  $self->{deep}=scalar(keys(%zpoints));

  #for my $key (sort{$a+0 <=> $b+0}keys(%{$self->{niveaulist}}))
  #{ print "$key = $self->{niveaulist}->{$key}\n"; }
  @{$self->{matrix}}=reverse(@{$self->{matrix}});

  # schrittweite bestimmen
  $self->{stepping_width}=$self->{real_width}/$self->{width};
  $self->{stepping_height}=$self->{real_height}/$self->{height};
  $self->{stepping_deep}=($self->{real_max_deep}-$self->{real_min_deep})/$self->{deep};

  return 1;
}

########################################################################
# Private Klassen_methoden #############################################
########################################################################

# funktioniert nur unter Linux, Unix, MacOS >= 10
# unter Windows macht das Probleme
# der Pfadtrenner ist hier "\"...
sub __base_path
{
  my $file=shift;
  (my $dir=$file)=~s![^/]+$!!;
  $dir=~s!/$!!;
  return $dir;
}

1;
